@@ -44,7 +44,15 @@
 ;; (elserv-publish (elserv-find-process 8080) "/"
 ;;                 :string "Hello World."
 ;;                 :content-type "text/plain")
-;; elserv-demo.el is a sample demo server.
+;;
+;; or write following lines in your .emacs.
+;;
+;; (autoload elserv-start "elserv" nil t)
+;; (add-hook 'elserv-start-hook
+;;           '(lambda ()
+;;              (elserv-publish (elserv-find-process) "/"
+;;                              :string "Hello World."
+;;                              :content-type "text/plain")))
 
 ;;; History:
 ;;
@@ -62,56 +70,115 @@
   (require 'cl)
   (require 'static))
 
+(eval-and-compile
+  (autoload 'elserv-autoindex "elserv-autoindex" nil t))
+
 (product-provide 'elserv
   (product-define "Elserv" nil
-		  '(0 3 1)
-		  "Miss Me Blind"))
+		  '(0 4 0)
+		  "Never Surrender"))
 
-(defvar elserv-default-port 8000
-  "*Default port number for Elserv.")
+(defgroup elserv nil
+  "Elserv -- Yet another HTTP server on Emacsen."
+  :group 'hypermedia)
 
-(defvar elserv-program-name nil
+(defcustom elserv-default-port 8000
+  "*Default port number for Elserv."
+  :type 'integer
+  :group 'elserv)
+
+(defcustom elserv-program-name nil
   "*If non-nil, it is invoked as a command.
-`elserv-daemon-name' is passed as first argument.")
+`elserv-daemon-name' is passed as first argument."
+  :type '(choice (symbol :tag "Direct" nil)
+		 (string :tag "Program Name"))
+  :group 'elserv)
 
-(defvar elserv-daemon-name "elservd"
-  "*Program name for Elserv daemon process.")
+(defcustom elserv-daemon-name (if (fboundp 'locate-data-directory)
+				  (expand-file-name
+				   "elservd"
+				   (locate-data-directory "elserv"))
+				"elservd")
+  "*Program name for Elserv daemon process."
+  :type 'string
+  :group 'elserv)
 
-(defvar elserv-publish-hash-length 31
-  "*Length of publish hash.")
+(defcustom elserv-publish-hash-length 31
+  "*Length of publish hash."
+  :type 'integer
+  :group 'elserv)
 
-(defvar elserv-debug t
-  "*If non-nil, request string is inserted to the debug buffer.")
+(defcustom elserv-debug nil
+  "*If non-nil, request string is inserted to the debug buffer."
+  :type 'boolean
+  :group 'elserv)
 
-(defvar elserv-directory-index-file "index.html"
-  "*Index file for the directory.")
+(defcustom elserv-directory-index-file "index.html"
+  "*Index file name for the directory."
+  :type 'string
+  :group 'elserv)
 
-(defvar elserv-keep-alive t
+(defcustom elserv-directory-autoindex t
+  "*If Non-nil and directory has no index file, generate html index in the
+directory."
+  :type 'boolean
+  :group 'elserv)
+
+(defcustom elserv-keep-alive t
   "*Non-nil enable persistent connections.
-\(more than one request per connection\).")
+\(more than one request per connection\)."
+  :type 'boolean
+  :group 'elserv)
 
-(defvar elserv-max-keep-alive-requests 100
+(defcustom elserv-max-keep-alive-requests 100
   "*The maximum number of requests to allow during a persistent connection.
 Set to nil to allow an unlimited amount.
-We recommend you leave this number high, for maximum performance.")
+We recommend you leave this number high, for maximum performance."
+  :type 'integer
+  :group 'elserv)
 
-(defvar elserv-keep-alive-timeout 15
-  "*Number of seconds to wait for the next request on the same connection.")
+(defcustom elserv-keep-alive-timeout 15
+  "*Number of seconds to wait for the next request on the same connection."
+  :type 'integer
+  :group 'elserv)
 
-(defvar elserv-identity-check nil
+(defcustom elserv-identity-check nil
   "*Non-nil enables RFC1413-compliant logging.
-\(logging of the remote user name for each connection\)")
+\(logging of the remote user name for each connection\)"
+  :type 'boolean
+  :group 'elserv)
 
-(defvar elserv-max-clients 20
+(defcustom elserv-max-clients 20
   "*Non-nil limits the number of clients who can simultaneously connect.
-If this limit is ever reached, clients will be LOCKED OUT.")
+If this limit is ever reached, clients will be LOCKED OUT."
+  :type 'integer
+  :group 'elserv)
 
-(defvar elserv-access-log-file nil
-  "*If file name is specified, access log is saved to the file.")
+(defcustom elserv-access-log-file nil
+  "*If file name is specified, access log is saved to the file."
+  :type 'file
+  :group 'elserv)
+
+(defcustom elserv-access-log-max-size 50000
+  "*Max size of access log file."
+  :type 'integer
+  :group 'elserv)
+
+(defcustom elserv-icon-path (if (fboundp 'locate-data-directory)
+				(locate-data-directory "elserv")
+			      (let ((icons (expand-file-name "elserv/icons/"
+							     data-directory)))
+				(if (file-directory-p icons)
+				    icons)))
+  "*Icon directory path."
+  :type 'directory
+  :group 'elserv)
+
+(defcustom elserv-icon-publish-path "/icons"
+  "*Path to publish an icon directory specified by `elserv-icon-path'."
+  :type 'string
+  :group 'elserv)
 
-(defvar elserv-access-log-max-size 50000
-  "*Max size of access log file.")
-  
 (defconst elserv-url-unreserved-chars
   '(?a ?b ?c ?d ?e ?f ?g ?h ?i ?j ?k ?l ?m
        ?n ?o ?p ?q ?r ?s ?t ?u ?v ?w ?x ?y ?z
@@ -122,10 +189,10 @@ If this limit is ever reached, clients will be LOCKED OUT.")
 
 (defconst elserv-http-version "HTTP/1.1")
 
-(defvar elserv-server-eol "\r\n"
-  "The EOL string sent from the server.")
+(defconst elserv-server-eol "\r\n"
+  "The end-of-line string sent from the server.")
 
-(defvar elserv-client-eor "\r\n\r\n"
+(defconst elserv-client-eor "\r\n\r\n"
   "The end-of-request string sent from the elservd.")
 
 (defvar elserv-buffer-publish-hash nil)
@@ -334,14 +401,17 @@ credentials (e.g., bad password), or your browser doesn't understand how to supp
       (elserv-version) "</body></html>"))
     result))
 
-(defun elserv-make-redirect (where)
-  "Redirect the client REQ to new location WHERE."
-  (elserv-make-result 'elserv-moved-permanently
-		      (list 'location where
-			    'content-type "text/html"
-			    'uri where)
-		      "<html><head><title>Moved permanently</title></head>
-<body><h1>Moved permanently</h1>This Page is moved permanently."))
+(defun elserv-make-redirect (result where)
+  "Make RESULT as a redirect to new location WHERE."
+  (elserv-set-result-code result 'elserv-moved-permanently)
+  (elserv-set-result-header result
+			    (list 'location where
+				  'content-type "text/html"
+				  'uri where))
+  (elserv-set-result-body result
+			  "<html><head><title>Moved permanently</title></head>
+<body><h1>Moved permanently</h1>This Page is moved permanently.</body>")
+  result)
 
 (defun elserv-version (&optional arg)
   "Return Elserv version.
@@ -668,6 +738,7 @@ Return server process object."
     (get-buffer-create (concat "*Log of elserv*"
 			       (number-to-string
 				(elserv-process-port process))))
+    (run-hooks 'elserv-start-hook)
     process))
 
 (defun elserv-process-port (process)
@@ -846,7 +917,7 @@ Otherwise, RESULT is set as authenticated and return nil."
 PROCESS is the server process of Elserv.
 PATH is the requested path.
 Rest of arguments ARGS are plist of the form (:ATTR1 VAL1 :ATTR2 VAL2 ...)."
-  (let (data set-auth auth predicate host)
+  (let (data set-auth auth predicate host doc)
     ;; Virtual host.
     (if (setq host (plist-get args :host))
 	(setq path (concat host path)))
@@ -857,18 +928,21 @@ Rest of arguments ARGS are plist of the form (:ATTR1 VAL1 :ATTR2 VAL2 ...)."
 		    :realm (plist-get set-auth :realm)
 		    :users (plist-get set-auth :users))))
       (setq predicate (elserv-make-predicate-from-plist args))
+      (setq doc (plist-get args :description))
       (cond
        ((setq data (plist-get args :directory)) ; directory is set.
 	(set (intern path elserv-buffer-publish-hash)
-	     (list 'elserv-service-directory auth predicate data)))
+	     (list 'elserv-service-directory
+		   doc auth predicate data)))
        ((setq data (plist-get args :string))    ; string is set.
 	(set (intern path elserv-buffer-publish-hash)
 	     (list 'elserv-service-string
-		   auth predicate data
+		   doc auth predicate data
 		   (plist-get args :content-type))))
        ((setq data (plist-get args :function))   ; handler is set.
 	(set (intern path elserv-buffer-publish-hash)
-	     (nconc (list 'elserv-service-function auth predicate data
+	     (nconc (list 'elserv-service-function
+			  doc auth predicate data
 			  (plist-get args :content-type)))))))))
 
 (defun elserv-unpublish (process path)
@@ -878,28 +952,28 @@ PATH is the requested path."
   (with-current-buffer (process-buffer process)
     (unintern path elserv-buffer-publish-hash)))
 
-(defsubst elserv-call-service-function-maybe (ppath path host request)
+(defsubst elserv-execute-service-maybe (ppath path host request)
   "Call service function for PPATH, PATH, HOST and REQUEST, if registered.
 Return result structure. If function is not registered, return nil."
   (let (sym func)
-    (when (and (or (boundp (setq sym (intern (concat host ppath)
-					     elserv-buffer-publish-hash)))
-		   (boundp (setq sym (intern ppath
-					     elserv-buffer-publish-hash))))
+    (when (and (or (setq sym (intern-soft 
+			      (concat host ppath)
+			      elserv-buffer-publish-hash))
+		   (setq sym (intern-soft
+			      ppath
+			      elserv-buffer-publish-hash)))
+	       (boundp sym)
 	       (setq func (append (symbol-value sym)
 				  (list path ppath request))))
       (apply (car func) (cdr func)))))
 
 (defun elserv-parse-path (path)
   "Return a reversed list of substrings of PATH which are separated by '/'."
-  (let ((start (if (eq (aref path 0) ?/) 1 0))
-	parts)
+  (let ((start 0) parts)
     (while (string-match "/" path start)
       (setq parts (cons (substring path start (match-beginning 0)) parts)
 	    start (match-end 0)))
-    (if (eq (length path) start)
-	parts
-      (cons (substring path start) parts))))
+    (cons (substring path start) parts)))
 
 (defun elserv-service (process path request)
   "Provide a service.
@@ -907,7 +981,7 @@ PROCESS is the server process of Elserv.
 PATH is the requested path string.
 REQUEST is the request structure."
   (let ((host (plist-get request 'host))
-	path-list ppath result)
+	path-list ppath rpath result)
     ;; absolute URI.
     (when (string-match "^http://\\([^/]+\\)\\(/\\)" path)
       (setq host (substring path (match-beginning 1) (match-end 1))
@@ -915,18 +989,19 @@ REQUEST is the request structure."
     (setq path-list (elserv-parse-path path))
     (with-current-buffer (process-buffer process)
       (while path-list
-	(setq ppath (concat "/" (mapconcat 'identity
-					   (reverse path-list) "/")))
-	(if (setq result (elserv-call-service-function-maybe
-			  ppath (substring path (length ppath))
+	(setq ppath (concat (mapconcat 'identity
+				       (reverse path-list) "/"))
+	      rpath (substring path (length ppath)))
+	(when (eq (length ppath) 0)
+	  (setq ppath "/"))
+	(when (string= ppath "/")
+	  (setq rpath path))
+	(if (setq result (elserv-execute-service-maybe
+			  ppath rpath
 			  host request))
 	    (setq path-list nil))
 	(setq path-list (cdr path-list)))
       (or result
-	  ;; Try root.
-	  (elserv-call-service-function-maybe "/"
-					      (substring path 1)
-					      host request)
 	  (signal 'elserv-file-not-found path)))))
 
 (defun elserv-parse-accept-language (string)
@@ -953,8 +1028,9 @@ REQUEST is the request structure."
 	      filename)))
     filename))
 
-(defun elserv-service-directory (auth predicate root path ppath request)
+(defun elserv-service-directory (doc auth predicate root path ppath request)
   "Service a directory.
+DOC is the documentation of the service.
 AUTH is the autenticator plist.
 PREDICATE is the predicate to check a request.
 ROOT is the top directory recorded by `elserv-publish'.
@@ -965,19 +1041,20 @@ REQUEST is the request structure (plist)."
     (or (elserv-check-predicate request predicate)
 	(elserv-authenticate request auth result)
 	(let (filename realfile attr mime-type)
-	  (setq filename (concat root (unless (eq (length path) 0) "/") path))
+	  (setq filename (concat root path))
 	  (setq path (elserv-url-decode-string path))
 	  (when (string-match "\\.\\." path)
 	    (signal 'elserv-forbidden (concat root path)))
-	  (if (or (and (string= ppath "/") (eq (length path) 0))
-		  (zerop (length (file-name-nondirectory filename))))
+	  (if (zerop (length (file-name-nondirectory filename)))
 	      (setq filename (expand-file-name
-			      elserv-directory-index-file
+ 			      elserv-directory-index-file
 			      filename)))
 	  (cond ((file-directory-p filename)
 		 (elserv-make-redirect
+		  result
 		  (concat "http://" (plist-get request 'host)
-			  ppath (unless (eq (length path) 0) "/") path "/")))
+			  (unless (string= ppath "/") ppath)
+			  path "/")))
 		((setq realfile
 		       (elserv-find-file filename (plist-get 
 						   request
@@ -997,11 +1074,23 @@ REQUEST is the request structure (plist)."
 					    realfile)
 					   (buffer-string)))
 		 result)
-		(t (signal 'elserv-file-not-found path)))))))
+		((and elserv-directory-autoindex
+		      (string= elserv-directory-index-file
+			       (file-name-nondirectory filename)))
+		 (elserv-autoindex
+		  result
+		  (plist-get request 'host) (concat
+					     (unless (string= ppath "/") ppath)
+					     path)
+		  (file-name-directory filename)))
+		(t (signal 'elserv-file-not-found
+			   (concat (unless (string= ppath "/") ppath)
+				   path))))))))
 
-(defun elserv-service-string (auth predicate string content-type path ppath
-				   request)
+(defun elserv-service-string (doc auth predicate string content-type path ppath
+				  request)
   "Service a string.
+DOC is the documentation of the service.
 AUTH is the autenticator plist.
 PREDICATE is the predicate to check a request.
 STRING is the content string recorded by `elserv-publish'.
@@ -1019,9 +1108,10 @@ REQUEST is the request structure (plist)."
 	  (elserv-set-result-body result string)
 	  result))))
 
-(defun elserv-service-function (auth predicate function
+(defun elserv-service-function (doc auth predicate function
 				     content-type path ppath request)
   "Service by a function.
+DOC is the documentation of the service.
 AUTH is the autenticator plist.
 PREDICATE is the predicate to check a request.
 FUNCTION is the symbol of the function registered.
@@ -1034,21 +1124,85 @@ REQUEST is the request structure (plist)."
 	(elserv-authenticate request auth result)
 	(progn
 	  (funcall function result
-		   (elserv-url-decode-string path) ppath request)
-	  (elserv-set-result-code result 'elserv-ok)
-	  (unless (plist-get (elserv-result-header result) 'content-type)
-	    (elserv-set-result-header result
-				      (append
-				       (elserv-result-header result)
- 				       `(content-type ,(or content-type
-							   "text/plain")))))
+		   (elserv-url-decode-string path)
+		   ppath request)
+	  (unless (elserv-result-code result)
+	    (elserv-set-result-code result 'elserv-ok)
+	    (unless (plist-get (elserv-result-header result) 'content-type)
+	      (elserv-set-result-header result
+					(append
+					 (elserv-result-header result)
+					 `(content-type ,(or content-type
+							     "text/plain"))))))
 	  result))))
 
+(defun elserv-package-publish (process path name)
+  "Publish package.
+PROCESS is the server process of Elserv.
+PATH is the path to publish.
+NAME is the name of the package to publish."
+  (require (intern (concat "es-" name)))
+  (let ((sym (intern (concat "elserv-" name "-publish"))))
+    (if (fboundp sym)
+	(funcall sym process path)
+      (error "Cannot publish as package: %s." name))))
+
 (defun elserv-publish-default (process)
-  "Publish tiny default page for PROCESS."
-  (elserv-publish process "/"
-		  :string "Hello World."
-		  :content-type "text/html"))
+  "Publish default pages for PROCESS."
+  ;; Publish monitor.
+  (elserv-package-publish process "/" "monitor")
+  (elserv-package-publish process "/monitor" "monitor")
+  ;; Publish icons.
+  (if (and elserv-icon-path
+	   (file-directory-p elserv-icon-path))
+      (elserv-publish process elserv-icon-publish-path
+		      :directory elserv-icon-path)))
+
+;;; Utils
+
+(defun elserv-replace-in-string (str regexp newtext &optional literal)
+  "Replace all matches in STR for REGEXP with NEWTEXT string.
+And returns the new string.
+Optional LITERAL non-nil means do a literal replacement.
+Otherwise treat \\ in NEWTEXT string as special:
+  \\& means substitute original matched text,
+  \\N means substitute match for \(...\) number N,
+  \\\\ means insert one \\."
+  (let ((rtn-str "")
+	(start 0)
+	(special)
+	match prev-start)
+    (while (setq match (string-match regexp str start))
+      (setq prev-start start
+	    start (match-end 0)
+	    rtn-str
+	    (concat
+	     rtn-str
+	     (substring str prev-start match)
+	     (cond (literal newtext)
+		   (t (mapconcat
+		       (function
+			(lambda (c)
+			  (if special
+			      (progn
+				(setq special nil)
+				(cond ((eq c ?\\) "\\")
+				      ((eq c ?&)
+				       (substring str (match-beginning 0)
+						  (match-end 0)))
+				      ((and (>= c ?0) (<= c ?9))
+				       (if (> c (+ ?0 (length
+						       (match-data))))
+					   ;; Invalid match num
+					   (error "Invalid match num: %c" c)
+					 (setq c (- c ?0))
+					 (substring str (match-beginning c)
+						    (match-end c))))
+				      (t (char-to-string c))))
+			    (if (eq c ?\\) (progn (setq special t) nil)
+			      (char-to-string c)))))
+		       newtext ""))))))
+    (concat rtn-str (substring str start))))
 
 (provide 'elserv)
 
